/*
====================================================================================================
    __  __    __  ___    ____    ____ 
    \ \/ /   /  |/  /   /  _/   / __ \
     \  /   / /|_/ /    / /    / /_/ /
     / /   / /  / /   _/ /    / _, _/ 
    /_/   /_/  /_/   /___/   /_/ |_|  

    YMIR SHADERS

    Copyright (C) 2020 Jessie, RRe36, Niemand

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://github.com/ymir-graphics/ymir-releases/blob/master/LICENSE

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/


/*
planar clouds layer 0, Altostratus clouds
*/

const float pc0_size    = 0.0009;
const float pc0_density_mult = 0.08;
const float pc0_highEdge = pcloud0_alt+pcloud0_depth;

float pc0_get_coverage(vec3 pos) {
    float erode_low     = 1.0-sstep(pos.y, pcloud0_alt, pcloud0_alt+pcloud0_depth*0.12);
    float erode_high    = sstep(pos.y, pcloud0_alt+pcloud0_depth*0.3, pc0_highEdge);
    float fade_low      = sstep(pos.y, pcloud0_alt, pcloud0_alt+pcloud0_depth*0.07);
    float fade_high     = 1.0-sstep(pos.y, pcloud0_alt+pcloud0_depth*0.9, pc0_highEdge);

    vec3 wind       = vec3(vcloud_tick*0.66, 0.0, vcloud_tick*0.34);

    pos            *= pc0_size;
    pos            += wind*0.6;

    float local_cov = noise_2d(pos.xz*0.00011+wind.xz*0.000018).x;

    if (local_cov < 0.02) return 0.0;

    float shape     = local_cov;
        shape      *= noise_2d(pos.xz*0.001+wind.xz*0.001).x*0.3+0.7;
        shape      -= 0.34+daytime.y*0.12+daytime.z*0.09+daytime.w*0.06;

        shape      *= fade_low;
        shape      *= fade_high;
        shape      -= erode_low*0.16*saturate(shape*2.0);
        shape      -= erode_high*0.16*saturate(shape*2.0);

    return max(shape, 0.0);
}

float pc0_get_shape(vec3 pos, float coverage) {
    vec3 wind       = vec3(vcloud_tick*0.66, 0.0, vcloud_tick*0.34);

    pos            *= pc0_size;
    pos            += wind*0.6;

    float shape     = coverage;
    
    float div       = 0.0;
    float noise     = 0.0;

    vec3 s1         = noise_3d(pos*0.03+coverage*0.3);

        noise      += s1.x;          div += 1.0;
        noise      += s1.y*0.75*saturate(1.0-coverage*0.7);     div += 0.75;

        noise      /= div;

        shape      -= (1.0-noise)*0.24*saturate(1.0-coverage*0.5);

    vec3 pos2       = pos*1.3 + coverage*0.15+noise*0.15;

    float noise2    = fbm(pos2, wind*0.2, 0.5, 3.0, 4)*1.0;

        shape      -= (1.0-noise2)*0.07;

        shape       = max(shape, 0.0);
        //float c     = (daytime.y + daytime.z * 0.5)*0.6;
        //shape      *= shape * c + (1.0-c);

    return shape;
}

float pc0_directLight_od(in vec3 pos) {
    vec3 direction = vc_lightvec;
    const int steps = 4;

    const float basestep = mix(120.0, 12.0, saturate(float(steps-1)/4.0));
    const float exponent = mix(2.0, 1.1, saturate(float(steps-8)/8.0));

    float stepsize  = basestep;
    float prev_step = stepsize;

    float od = 0.0;
    for(int i = 0; i < steps; ++i, pos += direction * stepsize) {
        if(pos.y > pc0_highEdge || pos.y < pcloud0_alt) break;
        prev_step = stepsize;
        stepsize *= exponent;

        float coverage  = pc0_get_coverage(pos);
        if(coverage <= 0.0) { continue; }
        float attenCoeff = pc0_get_shape(pos, coverage);
        od += attenCoeff*prev_step;
    } 

    return od;
}

float pc0_skyLight_od(in vec3 pos) {
    const vec3 direction = vec3(0, 1, 0);
    const int steps = 2;

    float stepsize = (pcloud0_depth / steps)*0.6;
        stepsize  *= linStep(pos.y, pcloud0_alt, pc0_highEdge)*0.9+0.1;

    vec3 increment = direction * stepsize;

    float od = 0.0;
    for(int i = 0; i < steps; ++i, pos += increment) {
        if(pos.y > pc0_highEdge) break;
        float coverage  = pc0_get_coverage(pos);
        if(coverage <= 0.0) { continue; }
        float attenCoeff = pc0_get_shape(pos, coverage);
        od += attenCoeff*stepsize;
    } 

    return od;
}

#define pc0_clip 150000.0

void pc0_compute_clouds(inout vec3 scenecol, vec3 wvec, float dither, inout float layer_transmittance) {
    vec2 sphere     = rsi(vec3(0.0, planetRadius+eyeAltitude, 0.0), wvec, planetRadius);
    bool visible    = !((eyeAltitude<pcloud0_alt && sphere.y>0.0) || (eyeAltitude>pc0_highEdge && wvec.y>0.0));

    if (visible) {
        const float pc_midAlt   = pcloud0_alt+pcloud0_depth*0.2;

        vec2 s      = rsi(vec3(0.0, planetRadius+eyeAltitude, 0.0), wvec, planetRadius+pc_midAlt);

        float sdist = eyeAltitude>pc_midAlt ? s.x : s.y;

        vec3 spos   = wvec*sdist;

        spos        = planetCurvePosition(spos);

        vec3 rpos   = spos+cameraPosition+normalize(spos)*dither*180.0;

        float rlength = pcloud0_depth;

        vec2 scatter = vec2(0.0);
        float transmittance = 1.0;
        float od = 0.0;
        float fade  = 0.0;
        float fdist = pc0_clip;

        float vdotl = dot(wvec, vc_lightvec);
        float pfade = saturate(cornettShanksMiePhase(vdotl, 0.5));       //this is for fading the powder

        vec3 sunlight   = lightColor[0];
        vec3 skylight   = lightColor[1]*0.25;

        const float sigma_a = 1.00;         //absorption coeff
        const float sigma_s = 0.05;         //scattering coeff, can technically be assumed to be sigma_t since the albedo is close to 1.0
        const float sigma_t = 0.05;         //extinction coeff, 0.05-0.12 for cumulus, 0.04-0.06 for stratus

        float dist  = distance(rpos, cameraPosition);
        float dfade = saturate(dist/pc0_clip);
        if ((1.0-dfade)>0.01) {

            float coverage  = pc0_get_coverage(rpos);
            if (coverage>0.01) {

                float density   = pc0_get_shape(rpos, coverage);
                if (density>0.0) {
                    fdist = dist;
                    fade = linStep(dfade, 0.75, 0.999);

                    float extinction = density * sigma_t;
                    float stept     = expf(-extinction*rlength);
                    float integral  = (1.0 - 1.0*stept) / sigma_t;

                    vec2 result_s   = vec2(0.0);

                    float directod  = pc0_directLight_od(rpos)*sigma_a;
                    float skyod     = pc0_skyLight_od(rpos)*sigma_a;

                    float powder    = 1.0 - expf(-density * 30.0);
                    float dpowder   = min(powder, 1.0-expf(-directod*2.0));    //because sun od based powder simply adds this little "extra", using min here gives a much more pleasurable result
                        dpowder     = mix(dpowder, 1.0, pfade);

                    for (int j = 0; j<3; ++j) {
                        float n     = float(j);

                        float s_d   = sigma_s * pow(0.5, n);    //scatter derivate
                        float t_d   = sigma_t * pow(0.5, n);    //transmittance/attentuation derivate
                        float phase = cloud_phase(vdotl, pow(0.5, n));  //phase derivate

                        result_s.x += expf(-directod*t_d) * phase * dpowder * s_d;
                        result_s.y += expf(-skyod*t_d) * powder * s_d;
                    }

                    scatter    += result_s * integral * transmittance;

                    transmittance *= stept;
                }
            }
        }

        //transmittance = linStep(transmittance, pcloud0_thresh, 1.0); //remap transmittance so that the threshold is zero

        vec3 color      = (sunlight*scatter.x) + (skylight*scatter.y);

        vec3 atmosAirmass = calculateAtmosphereDensity(length(wvec) * wvec.y * 0.5 + eyeAltitude + planetRadius);
        vec3 atmosStepOD = extinctionCoefficients * atmosAirmass;

        vec3 atmosTransmittance = expf(-atmosStepOD * fdist);

        fade    = saturate(1.0-fade);
        fade    = sqr(fade);

        scenecol        = mix(scenecol, scenecol * transmittance + color, atmosTransmittance*fade*layer_transmittance);

        layer_transmittance *= transmittance;
    }
}