
/*
====================================================================================================
    __  __    __  ___    ____    ____ 
    \ \/ /   /  |/  /   /  _/   / __ \
     \  /   / /|_/ /    / /    / /_/ /
     / /   / /  / /   _/ /    / _, _/ 
    /_/   /_/  /_/   /___/   /_/ |_|  

    YMIR SHADERS

    Copyright (C) 2020 Jessie, RRe36, Niemand

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://github.com/ymir-graphics/ymir-releases/blob/master/LICENSE

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/


#define HBAO_RADIUS 1.5 //[0.25 0.5 0.75 1.0 1.25 1.5 1.75 2.0 2.25 2.5 2.75 3.0 3.5 4.0 4.5 5.0 5.5 6.0 6.5 7.0 7.5 8.0 8.5 9.0 9.5 10.0] Changes the radius of HBAO.
#define HBAO_ANGLE_SAMPLES 6 //[1 2 4 6 8 10 12 16] Sets the angle samples.
#define HBAO_HORIZON_DIRECTIONS 6 //[1 2 4 6 8 10 12 16] Set the horion directions.

vec3 calculate_base_horizon_vector(vec3 Po, vec3 Td, vec3 L, vec3 N, float LdotN) {
	vec3 negPoLd = Td - Po;
	float D = -dot(negPoLd, N) / LdotN;
	return normalize(D * L + negPoLd);
}

float calculate_horizon_angle(vec3 position, vec2 screenCoord, vec3 horizonDirection, vec3 viewDirection,
	    vec3 normal, float NdotV, float sampleOffset) {

	vec3 horizonVector = calculate_base_horizon_vector(position, horizonDirection, viewDirection, normal, NdotV);
	float cosHorizonAngle = clamp(dot(horizonVector, viewDirection), -1.0, 1.0);

	for (int i = 0; i < HBAO_ANGLE_SAMPLES; ++i) {
		float sampleDistance2D = pow(float(i) / float(HBAO_ANGLE_SAMPLES) + sampleOffset, 2.0);
		vec2 sampleCoord = horizonDirection.xy * sampleDistance2D + screenCoord;

		if (clamp(sampleCoord, 0.0, 1.0) != sampleCoord) { break; }

		vec3 samplePosition   = vec3(sampleCoord, texture(depthtex0, sampleCoord).r);
        if (samplePosition.z < 0.56) continue;  //the hand causes bugs, so we ignore it
		//if (samplePosition.z == 1.0) continue;
		     samplePosition.z = 1e-3 * (1.0 - samplePosition.z) + samplePosition.z;

		samplePosition = screenSpaceToViewSpace(samplePosition, gbufferProjectionInverse);

		vec3  sampleVector          = samplePosition - position;
		float sampleDistanceSquared = dot(sampleVector, sampleVector);
		vec3  sampleDirection       = sampleVector * inversesqrt(sampleDistanceSquared);

		if (sampleDistanceSquared > HBAO_RADIUS * HBAO_RADIUS) { continue; }

		float cosSampleAngle = dot(viewDirection, sampleDirection);


		cosHorizonAngle = clamp(cosSampleAngle, cosHorizonAngle, 1.0);
	}

	return acos(cosHorizonAngle);
}

float compute_hbao(vec3 viewpos, vec2 coord, vec3 viewDir, vec3 viewNormal, float dither) {
	const float ditherSize = _sqr(32.0) * 16.0;
    dither = ditherSize * dither + 0.5;

    float nDotV     = dot(viewNormal, viewDir);

	vec2 norm = vec2(gbufferProjection[0].x, gbufferProjection[1].y) * ((-0.5 * HBAO_RADIUS) / viewpos.z);

	float result = 0.0;
	for (int i = 0; i < HBAO_HORIZON_DIRECTIONS; ++i) {
		float theta = (i * ditherSize + dither) * golden_angle;
		vec3 horizon_dir     = vec3(abs(sin(theta)), cos(theta), 0.0);
		     horizon_dir.xy *= norm;

		float sampleOffset = (i + dither / ditherSize) / (HBAO_ANGLE_SAMPLES * HBAO_HORIZON_DIRECTIONS);
		result += calculate_horizon_angle(viewpos, coord,  horizon_dir, viewDir, viewNormal, nDotV, sampleOffset);
		result += calculate_horizon_angle(viewpos, coord, -horizon_dir, viewDir, viewNormal, nDotV, sampleOffset);
	}

	return result / (HBAO_HORIZON_DIRECTIONS);
}