
/*
====================================================================================================
    __  __    __  ___    ____    ____ 
    \ \/ /   /  |/  /   /  _/   / __ \
     \  /   / /|_/ /    / /    / /_/ /
     / /   / /  / /   _/ /    / _, _/ 
    /_/   /_/  /_/   /___/   /_/ |_|  

    YMIR SHADERS

    Copyright (C) 2020 Jessie, RRe36, Niemand

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://github.com/ymir-graphics/ymir-releases/blob/master/LICENSE

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

vec3 orenNayar(in float roughness, in vec3 albedo, in vec3 viewDir, in vec3 lightDir, in vec3 normal) {
    /*
        This is from:
        http://www1.cs.columbia.edu/CAVE/publications/pdfs/Oren_CVPR93.pdf
	*/
    float nDotL = saturate(dot(normal, lightDir));
    float nDotV = saturate(dot(normal, viewDir));

    vec2 angles = acos(vec2(nDotL, nDotV));

    if(angles.x < angles.y) {
        angles.xy = angles.yx;
    }
    float cosA = dot(viewDir - nDotV * normal, lightDir - nDotL * normal);

    float C1 = 1.0 + (roughness / (roughness + 0.4)) * (2.0 / pi - 1.0);
    float C2 = 0.4 * (roughness / (roughness + 0.2));
    if(cosA > 0.0) {
        C2 = C2 * pow(sin(angles.x), 0.56 + (0.20 / roughness));
    } else {
        C2 = C2 * (pow(sin(angles.x), 0.56 + (0.20 / roughness)) - pow(2.0*angles.y/pi, 2.0+(1.0/2.0*roughness)));
    }
    float C3 = 0.11 * (roughness / (roughness + 0.2)) * pow(4.0 * angles.x * angles.y / piSquared, 1.55 + (1.0 / 1.55 * roughness));

    vec3 Lr1 = (albedo / pi) * nDotL * (C1 + (cosA * C2 * tan(angles.y)) + ((1.0 - abs(cosA)) * C3 * tan((angles.x + angles.y) / 2.0)));
    vec3 Lr2 = sqr(albedo) * nDotL * (0.045 * (roughness / (roughness + 0.2))) * (1.0 - cosA * pow(2.0 * angles.y / pi, 1.3 + (0.31 / roughness)));
    return saturate(Lr1 + Lr2);
}

vec3 orenNayarQualitative(in float roughness, in vec3 albedo, in vec3 viewDir, in vec3 lightDir, in vec3 normal) {
    /*
        This is from:
        http://www1.cs.columbia.edu/CAVE/publications/pdfs/Oren_CVPR93.pdf
	*/
    float nDotL = saturate(dot(normal, lightDir));
    float nDotV = saturate(dot(normal, viewDir));

    vec2 angles = acos(vec2(nDotL, nDotV));

	if(angles.x < angles.y) angles = angles.yx;
    float cosA = saturate(dot(normalize(viewDir - nDotV * normal), normalize(lightDir - nDotL * normal)));

    vec3 A = albedo * ((1.0 / pi) - 0.09 * (roughness / (roughness + 0.4)));
    vec3 B = albedo * (0.125 * (roughness /  (roughness + 0.18)));

    vec3 Lr1 = nDotL * (A + B * max(0.0, cosA) * sin(angles.x) * tan(angles.y));

    return saturate(Lr1);
}

vec3 lambertianDiffuse(vec3 albedo, vec3 normal, vec3 direction) {
    /*
        This is used for some optimized or approximated stuff where Oren-Nayar gives very little visual advantages
        but a big performance loss instead.
    */
    float lambert   = saturate(dot(normal, direction));

    return albedo * lambert;
}