/*
====================================================================================================
    __  __    __  ___    ____    ____ 
    \ \/ /   /  |/  /   /  _/   / __ \
     \  /   / /|_/ /    / /    / /_/ /
     / /   / /  / /   _/ /    / _, _/ 
    /_/   /_/  /_/   /___/   /_/ |_|  

    YMIR SHADERS

    Copyright (C) 2020 Jessie, RRe36, Niemand

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://github.com/ymir-graphics/ymir-releases/blob/master/LICENSE

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/




#include "/lib/universal.glsl"

/* ------ framebuffer attachment outputs ------ */
/*DRAWBUFFERS:312*/
layout(location = 0) out vec4 scenecol_data;
layout(location = 1) out vec4 auxiliary_data1;
layout(location = 2) out vec4 auxiliary_data2;

/* ------ inputs from vertex stage ------ */
in mat2x2 coord;

flat in int mat_id;

in float shadow_warp;
in float shadow_cd;
in float view_dist;
flat in float lightFlip;

in vec3 pos_view;
in vec3 shadow_pos;
in vec3 world_pos;
in vec3 scenePos;
flat in vec3 normal;

in vec4 tint;

#ifdef g_terrain
    flat in ivec4 pbr_val;
#endif

flat in mat3x3 lightColor;
flat in mat3x3 tbn;

/* ------ uniforms ------ */
uniform int frameCounter;

uniform float frameTimeCounter;

uniform vec2 viewSize;

uniform vec3 lightVector, lightVectorView;

uniform mat4 gbufferModelView;
uniform mat4 gbufferModelViewInverse;
uniform mat4 shadowModelView;

uniform sampler2D tex;
uniform sampler2D noisetex;
uniform sampler2D shadowtex0;
uniform sampler2D shadowtex1;
uniform sampler2D shadowcolor0;

#ifdef specular_resourcepack
    uniform sampler2D specular;
#endif

#ifdef normals_resourcepack
    uniform sampler2D normals;
#endif

#ifdef cloud_shadows
    uniform sampler2D gaux3;
#endif

/* ------ includes ------ */
#include "/lib/util/srgb.glsl"
#include "/lib/util/encoders.glsl"
#include "/lib/lighting/diffuse.glsl"
#include "/lib/fragment/dither.glsl"
#include "/lib/util/bicubic.glsl"

#ifdef normals_resourcepack
    #include "/lib/fragment/labPBR.glsl"
#endif

vec3 get_blocklight(vec3 lightcol, float lightmap) {
    lightmap = linStep(lightmap, rcp(24.0), 1.0);
    return pow5(lightmap) * lightcol;
}

vec3 get_lighting(vec3 scenecol, vec3 normals, vec3 viewpos, vec3 lvec, mat3x3 lightcol, vec2 lmap, float ao) {
    vec3 indirect_light = vec3(0.3, 0.6, 1.0) * 0.00026 * scenecol * (ao * 0.9 + 0.1);
    return indirect_light + scenecol * get_blocklight(lightcol[2], lmap.x);
}


float bayer2e(vec2 a) {
    a = floor(a);
    return fract( dot(a, vec2(.5, a.y * .75)) );
}

#define bayer4e(a) (bayer2e( .5*(a))*.25+bayer2e(a))

#define m vec3(31,63,31)
float encode3x8(vec3 a) {
    float dither = bayer4e(gl_FragCoord.xy);
    a += (dither-.5) / m;
    a = saturate(a);
    ivec3 b = ivec3(a*m);
    return float( b.r|(b.g<<5)|(b.b<<11) ) / 65535.;
}
#undef m

#ifdef g_terrain
void water_customcolor(inout vec4 scenecol) {
    const vec3 watercol     = vec3(1.0, 1.0, 1.0);
    //This should be 0.0, as having it any higher can cause noticeable shadows on the water, which should never be visible outside of the sun reflection and the water volume.
    const float waterlum    = 0.0;
    const float opacity     = 0.12;

    scenecol.rgb    = watercol*waterlum;
    scenecol.a      = opacity;
}

vec2 rotate_coord(vec2 pos, const float angle) {
    return vec2(cos(angle)*pos.x + sin(angle)*pos.y, 
                cos(angle)*pos.y - sin(angle)*pos.x);
}

vec3 noise_2d(vec2 pos) {
    return texture(noisetex, pos).xyz;
}


float water_noise(vec2 pos) {
    return sqr(1.0-textureBicubic(noisetex, pos/256.0).z);
}

mat2 rotate(float a) {
    vec2 m;
    m.x = sin(a);
    m.y = cos(a);
	return mat2(m.y, -m.x,  m.x, m.y);
}

const float gravity = 9.81;

float gernsterWaves(vec2 coord, float time, float waveSteepness, float waveAmplitude, float waveLength, vec2 waveDirection){
	float k = tau / waveLength;
	float w = sqrt(gravity * k);

	float x = w * time - k * dot(waveDirection, coord);
	float wave = sin(x) * 0.5 + 0.5;

	float h = waveAmplitude * pow(wave, waveSteepness);

	return h;
}

float calculateWaves(in vec2 position) {
	float amplitude = 0.1;
	float steepness = 0.7;
	float waveLength = 12.0;
	float noiseAmplitude = 0.75;

	vec2 dir = vec2(0.25, 1.0);

	float a = 2.6;

    float waveTime = frameTimeCounter * 1.5;
    float waves = 0.0;

	vec2 noisepos = position * 0.005;
    for(uint i = 0u; i < 8u; ++i) {
		a *= pi+i;
		vec2 noise = texture(noisetex, noisepos).rg * 2.0 - 1.0;
        waves -= gernsterWaves(position + (noise * 1.25), waveTime, steepness, amplitude, waveLength, dir);

		waveTime *= 1.02;

		amplitude *= 0.57;
		steepness *= 1.04;
		waveLength *= 0.70;
		dir *= rotate(a);
		noisepos *= rotate(a);
    }
    return waves - amplitude;
}

vec3 waterParallax(vec3 position, vec3 viewDir, in vec3 viewP) {
    const int numLayers = 16;
    float layerDepth = 1.0 / float(numLayers);
    float currentLayerDepth = 0.0;
    vec3 interval = inversesqrt(numLayers) * viewDir / abs(viewDir.y);

    vec3  currentPosition = vec3(0.0);
    float currentDepthMapValue = calculateWaves(position.xz) * 2.0;

    for(int i = 0; i < numLayers && currentLayerDepth > currentDepthMapValue; ++i) {
        currentPosition = (currentPosition.y - currentDepthMapValue) * interval + currentPosition;
        currentDepthMapValue = calculateWaves(currentPosition.xz + position.xz) * 2.0; 
        currentLayerDepth -= layerDepth;
        currentLayerDepth /= 2.0;
    }

    return vec3(1, 0, 1) * currentPosition + position;
}

vec3 water_normal(in vec2 position, in float dist) {
	vec2 diffs = vec2(calculateWaves(position + vec2(dist, -dist)), 
                      calculateWaves(position + vec2(-dist, dist)))
        - calculateWaves(position - vec2(dist, dist));
	vec3 wavenormal = vec3(-2.0 * dist, 4.0 * dist * dist, -2.0 * (dist * dist + dist));
	wavenormal.xz *= diffs;

	return normalize(tbn * wavenormal.xzy);
}
#endif

void main() {
    #ifdef hide_terrain
        discard;
    #endif

    scenecol_data            = gtex(tex);
    if(scenecol_data.a < 0.1) discard;
    scenecol_data.rgb       *= tint.rgb;

    scenecol_data.rgb        = SRGBtoACEScg(scenecol_data.rgb);

    vec3 scenenormal         = normal;

    vec2 specularData        = vec2(0.0, 1.0);

    if(mat_id == 101) {
        #ifdef g_terrain
            specularData     = vec2(encode2x8(vec2(pbr_val.xy) / 255.0), 1.0);

            water_customcolor(scenecol_data);
            scenenormal      = water_normal(world_pos.xz, 0.01);
        #endif
    } else {
        #ifdef specular_resourcepack
            vec4 spectex     = gtex(specular);
            specularData     = vec2(encode2x8(spectex.rg), 1.0);

            #ifdef normals_resourcepack
                vec4 ntex    = gtex(normals);
                scenenormal  = decode_lab_nrm(ntex.rgb, tbn, scenecol_data.a);
            #endif
        #else
            #ifdef g_terrain
                specularData = vec2(encode2x8(vec2(pbr_val.xy) / 255.0), 1.0);
            #endif

            #ifdef normals_resourcepack
                vec4 ntex    = gtex(normals);
                scenenormal  = decode_nrm(ntex.rgb, tbn);
            #endif
        #endif
    }

    vec3 albedo              = sqrt(scenecol_data.rgb);   //make the most of the limited precision

    scenecol_data.rgb        = get_lighting(scenecol_data.rgb, normal, pos_view, lightVectorView, lightColor, coord[1], tint.a);
    scenecol_data            = make_drawbuffer(scenecol_data.rgb, saturate(scenecol_data.a));

    auxiliary_data1          = vec4(encodeNormal(scenenormal), encode2x8(coord[1]), specularData.y);
    auxiliary_data2          = vec4(encodeMatID16(mat_id), encode3x8(normalize(albedo)), specularData.x, 1.0);
}