/*
====================================================================================================
    __  __    __  ___    ____    ____ 
    \ \/ /   /  |/  /   /  _/   / __ \
     \  /   / /|_/ /    / /    / /_/ /
     / /   / /  / /   _/ /    / _, _/ 
    /_/   /_/  /_/   /___/   /_/ |_|  

    YMIR SHADERS

    Copyright (C) 2020 Jessie, RRe36, Niemand

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://github.com/ymir-graphics/ymir-releases/blob/master/LICENSE

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/



#include "/lib/universal.glsl"

const int shadowMapResolution = 2560;

/* ------ outputs ------ */
out mat2x2 coord;

flat out int mat_id;

out float shadow_warp;
out float shadow_cd;
out float view_dist;
flat out float lightFlip;

out vec3 pos_view;
out vec3 shadow_pos;
out vec3 world_pos;
out vec3 scenePos;
flat out vec3 normal;

out vec4 tint;

#ifdef g_terrain
    flat out ivec4 pbr_val;
#endif

flat out mat3x3 lightColor;
flat out mat3x3 tbn;

/* ------ uniforms & attributes ------ */
uniform float eyeAltitude;
uniform float wetness;

uniform vec2 viewSize;
uniform vec2 taaOffset;

uniform vec3 sunVector, sunVectorView;
uniform vec3 moonVector, moonVectorView;
uniform vec3 sunPosition, upPosition;
uniform vec3 cameraPosition;

uniform mat4 gbufferModelViewInverse;
uniform mat4 gbufferModelView;
uniform mat4 shadowModelViewInverse, shadowProjectionInverse;
uniform mat4 shadowModelView, shadowProjection;

uniform sampler2D gaux1;

attribute vec4 at_tangent;
attribute vec4 mc_Entity;

/* ------ includes ------ */
#include "/lib/util/blackbody.glsl"
#include "/lib/lighting/warp.glsl"
#include "/lib/atmosphere/constants.glsl"
#include "/lib/atmosphere/transmittance.glsl"
#include "/lib/atmosphere/project.glsl"

#ifdef g_terrain
    #include "/lib/fragment/labPBR.glsl"
#endif

#define transMAD(mat, v) (mat3x3(mat) * (v) + (mat)[3].xyz)

#define diagonal2(mat) vec2((mat)[0].x, (mat)[1].y)
#define diagonal3(mat) vec3(diagonal2(mat), (mat)[2].z)
#define diagonal4(mat) vec4(diagonal3(mat), (mat)[2].w)

#define projMAD3(mat, v) (diagonal3(mat) * (v) + (mat)[3].xyz)
#define viewMAD(m, v) (mat3(m) * (v) + (m)[3].xyz)

vec3 get_shadowcoord(vec3 viewpos, vec3 lvec, const float bias, out float warp, out float cd) {
    //we can probably optimize this more
    vec3 pos    = viewpos;
        pos    += vec3(bias)*lvec;
        pos     = viewMAD(gbufferModelViewInverse, pos);
        pos     = viewMAD(shadowModelView, pos);
        pos     = projMAD3(shadowProjection, pos);
        pos.z  *= shadowFarPlaneMult;

        warp    = 1.0;      //initialize this just for safety
        pos.xy  = warp_shadowmap(pos.xy, warp);

    vec3 p2     = pos;      //this is needed because otherwise the depth comparision is kinda fucked, but better than far plane issues
        p2.xy  *= warp;
        p2      = projMAD3(shadowProjectionInverse, p2);
        cd      = -p2.z;

    return pos*0.5+0.5;
}

const float shadow_bias = 0.022 * (2048.0 / shadowMapResolution);    //this probably need some more adjustment

void main() {
    coord[0]       = (gl_TextureMatrix[0] * gl_MultiTexCoord0).xy;
    coord[1]       = (gl_TextureMatrix[1] * gl_MultiTexCoord1).xy;
    coord[1].y     = linStep(coord[1].y, 1.0/16.0, 1.0);
    normal 	       = mat3(gbufferModelViewInverse) * normalize(gl_NormalMatrix*gl_Normal);
    tint           = gl_Color;

    vec3 tangent   = mat3(gbufferModelViewInverse) * normalize(gl_NormalMatrix*at_tangent.xyz);
    vec3 binormal  = mat3(gbufferModelViewInverse) * normalize(gl_NormalMatrix*cross(at_tangent.xyz, gl_Normal.xyz) * at_tangent.w);
    tbn            = mat3(tangent.x, binormal.x, normal.x,
                         tangent.y, binormal.y, normal.y,
                         tangent.z, binormal.z, normal.z);

    view_dist      = length(gl_ModelViewMatrix * gl_Vertex);
    // view_vec       = normalize(tbn * (gl_ModelViewMatrix * gl_Vertex).xyz);

    vec4 pos       = gl_Vertex;
    pos_view       = pos.xyz;

    pos            = transMAD(gl_ModelViewMatrix, pos.xyz).xyzz;

    shadow_pos     = get_shadowcoord(pos.xyz, sunVectorView, shadow_bias, shadow_warp, shadow_cd);

    pos            = transMAD(gbufferModelViewInverse, pos.xyz).xyzz;

    scenePos        = pos.xyz;

    world_pos      = pos.xyz+cameraPosition;

    pos            = transMAD(gbufferModelView, pos.xyz).xyzz;
    gl_Position    = pos.xyzz * diagonal4(gl_ProjectionMatrix) + vec4(0, 0, gl_ProjectionMatrix[3].z, 0);
    gl_Position.xy += taaOffset * gl_Position.w;

    //lighting stuffs
    lightColor[0]   = dot(sunPosition, upPosition) > 0.5 ? calculateAtmosphereTransmittance(sunVector, vec3(0.0, planetRadius + eyeAltitude, 0.0)) * sunIlluminance : calculateAtmosphereTransmittance(moonVector, vec3(0.0, planetRadius + eyeAltitude, 0.0)) * moonIlluminance;
    lightColor[1]   = texture(gaux1, projectSky(vec3(0.0, 1.0, 0.0))).rgb * pi;
    lightColor[1]   = mix(lightColor[1], vec3(v3avg(lightColor[1]) * 1.5), wetness * 0.6);
    lightColor[2]   = blackbody(blocklightTemp) * lux(2000) * blocklightMult;

    float lf       = dot(sunVector, vec3(0.0, 1.0, 0.0))*0.5+0.5;
    lf             = linStep(lf, 0.48, 0.499)*(1.0-linStep(lf, 0.501, 0.52));

    lightFlip     = saturate(1.0-lf);

    mat_id = 100;

    #ifdef g_terrain
        #if defined hardcoded_specular && !defined specular_resourcepack
            if (mc_Entity.x == 10103) {                     // glass
                pbr_val = ivec4(242, lab_f0(0.042), 0, 0);
            } else if (mc_Entity.x == 10104) {              // ice
                pbr_val = ivec4(230, lab_f0(0.018), 0, 0);
            } else if (mc_Entity.x == 10105) {              // slime/honey
                pbr_val = ivec4(242, lab_f0(0.037), 0, 0);
            } else {
                pbr_val = ivec4(0, lab_f0(0.08), 0, 0);
            }
        #else 
            pbr_val = ivec4(0, lab_f0(0.08), 0, 0);
        #endif

        if (mc_Entity.x == 10001) {                         // water
            mat_id = 101;
            pbr_val = ivec4(255, lab_f0(0.02), 0, 0);
        }
    #endif
}