#version 450 compatibility


/*
====================================================================================================
    __  __    __  ___    ____    ____ 
    \ \/ /   /  |/  /   /  _/   / __ \
     \  /   / /|_/ /    / /    / /_/ /
     / /   / /  / /   _/ /    / _, _/ 
    /_/   /_/  /_/   /___/   /_/ |_|  

    YMIR SHADERS

    Copyright (C) 2020 Jessie, RRe36, Niemand

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://github.com/ymir-graphics/ymir-releases/blob/master/LICENSE

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/



#include "/lib/universal.glsl"


/* ------ framebuffer attachment outputs ------ */
/*DRAWBUFFERS:0347*/
layout(location = 0) out vec4 scene_color;
layout(location = 1) out vec4 indirectLight_data;
layout(location = 2) out vec4 atmosphere_data;
layout(location = 3) out vec4 temporal_data;


/* ------ inputs from vertex stage ------ */
in vec2 coord;     //this is always the topmost variable since it's needed everywhere

flat in float cloudBackscattering;

flat in vec3 cloudLightvector;

flat in mat2x3 lightColor;


/* ------ uniforms ------ */
uniform int frameCounter;

uniform float eyeAltitude;
uniform float far, near;
uniform float frameTimeCounter;
uniform float wetness;

uniform vec2 pixelSize, viewSize;
uniform vec2 taaOffset;

uniform vec3 sunVector, moonVector, lightVector;
uniform vec3 cameraPosition, previousCameraPosition;

uniform mat4 gbufferModelViewInverse, gbufferModelView;
uniform mat4 gbufferProjectionInverse, gbufferProjection;
uniform mat4 gbufferPreviousModelView, gbufferPreviousProjection;

uniform sampler2D colortex0;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex4;
uniform sampler2D colortex6;
uniform sampler2D colortex7;

uniform sampler2D depthtex0;

uniform sampler2D noisetex;
uniform sampler3D depthtex2;


/* ------ includes ------ */
#include "/lib/util/encoders.glsl"
#include "/lib/atmosphere/constants.glsl"
#include "/lib/atmosphere/transmittance.glsl"
#include "/lib/atmosphere/phase.glsl"
#include "/lib/lighting/diffuse.glsl"
#include "/lib/atmosphere/sun_moon.glsl"
#include "/lib/atmosphere/project.glsl"
#include "/lib/fragment/dither.glsl"
#include "/lib/fragment/lowresCommon.glsl"
#define PD_3DNOISE
#include "/lib/fragment/noise.glsl"
#include "/lib/util/bicubic.glsl"
#include "/lib/util/transforms.glsl"


/* ------ functions ------ */

vec3 planetCurvePosition(in vec3 x) {
    return vec3(x.x, length(x + vec3(0.0, planetRadius, 0.0))-planetRadius, x.z);
}


/* ------ cloud reflections ------ */
vec3 projectSphere(vec2 coord) {
        coord  *= vec2(tau, pi);
    vec2 lon    = sincos(coord.x) * sin(coord.y);

    return vec3(lon.x, cos(coord.y), lon.y);
}

#define PD_CLOUDRENDER
#define PD_REFLECTION
#include "/lib/atmosphere/clouds/cloudShapeModule.glsl"

#include "/lib/atmosphere/clouds/cloudModuleReflect.glsl"


/* ------ indirect lighting filter ------ */
float depth_lin(float depth) {
    return (2.0*near) / (far+near-depth * (far-near));
}

#define PD_FILTERLOD 2.0
#include "/lib/fragment/waveletFilter.glsl"

vec3 reproject(vec3 sceneSpace) {
    vec3 prevScreenPos = cameraPosition - previousCameraPosition;
    prevScreenPos = sceneSpace + prevScreenPos;
    prevScreenPos = transMAD(gbufferPreviousModelView, prevScreenPos);
    prevScreenPos = transMAD(gbufferPreviousProjection, prevScreenPos) * (0.5 / -prevScreenPos.z) + 0.5;

    return prevScreenPos;
}


void main() {
    vec4 sceneCol   = stex(colortex0);
    float sceneDepth = stex(depthtex0).x;
    vec4 tex2       = stex(colortex2);
    int matID       = decodeMatID16(tex2.x);


    /* ------ sky projection ------ */
    if (!landMask(sceneDepth)) {
        vec3 viewPos    = screenSpaceToViewSpace(vec3(coord, sceneDepth), gbufferProjectionInverse);
        vec3 scenePos   = viewSpaceToSceneSpace(viewPos, gbufferModelViewInverse);
        vec3 worldDir   = normalize(scenePos);

        sceneCol.rgb    = textureBicubic(colortex4, projectSky(worldDir)).rgb;

        vec3 sunmoon    = physicalSun(worldDir) + physicalMoon(worldDir);
            sunmoon    *= calculateAtmosphereTransmittance(worldDir, vec3(0.0, planetRadius + eyeAltitude, 0.0), extinctionCoefficients, 8);
        
        sceneCol.rgb   += sunmoon;

        #ifdef vcloud_enabled
            const float cLOD    = sqrt(CLOUD_RENDER_LOD);
            vec2 cloudCoord     = coord * rcp(cLOD) + vec2(1.0 - rcp(cLOD), 0.0);
            vec4 cloud  = texture(colortex4, cloudCoord);

            sceneCol.rgb = sceneCol.rgb * cloud.a + cloud.rgb;
        #endif
    }

    scene_color     = make_drawbuffer(sceneCol);


    /* ------ sky reflection map ------*/
    atmosphere_data = stex(colortex4);

    const float cLOD    = sqrt(CLOUD_RENDER_LOD);
    vec2 scaledCoord    = (coord - vec2(1.0 - rcp(cLOD), 0.0)) * cLOD;
    if (clamp(scaledCoord, -0.001, 1.001) == scaledCoord) atmosphere_data = vec4(0.0, 0.0, 0.0, 1.0);   //clear area of clouds to preven issues when writing fog to it later

    vec2 skyrefCoord = (coord - vec2(0.5, 0.0)) * SKYREF_LOD;

    if (clamp(skyrefCoord, 0.0 - pixelSize * 5.0, 1.0 + pixelSize * 5.0) == skyrefCoord) {
        vec2 coord  = saturate(skyrefCoord);
        vec3 worldDir   = normalize(projectSphere(coord) * vec3(1.0, 1.0, -1.0));

        vec3 skycol     = texture(colortex4, projectSky(worldDir)).rgb;

        atmosphere_data.rgb = skycol;
        atmosphere_data.rgb *= linStep(worldDir.y, -0.1, -0.02) * 0.5 + 0.5;

        #ifdef cloud_reflections_enabled
            float dither    = dither_bluenoise_static();

            vec4 cloud      = computeCloudLayersReflect(worldDir, dither, skycol);


            atmosphere_data.rgb     = atmosphere_data.rgb * cloud.a + cloud.rgb;
        #endif
    }

    atmosphere_data     = clamp_drawbuffer(atmosphere_data);


    /* ------ indirect lighting ------ */
    indirectLight_data  = stex(colortex3);

    #if (defined ambient_occlusion_enabled || defined rsm_enabled)
        vec2 aoCoord    = coord * 2.0;
        
        if (clamp(aoCoord, 0.0 - pixelSize, 1.0 + pixelSize) == aoCoord) {
            aoCoord     = saturate(aoCoord);

            vec4 currentColor       = filter_3x3(colortex3, depthtex0, coord, 1);

            #ifdef rsm_enabled
                vec3 position   = vec3(aoCoord, texelFetch(depthtex0, ivec2(coord * viewSize * 2.0), 0).r);
                    position    = screenSpaceToViewSpace(position, gbufferProjectionInverse);
                    position    = viewSpaceToSceneSpace(position, gbufferModelViewInverse);

                float currentDistance   = length(position);

                vec3 reprojection   = reproject(position);

                bool offscreen      = saturate(reprojection) != reprojection;

                vec4 historyData    = texture(colortex7, reprojection.xy * 0.5);

                float distanceDelta = distance(historyData.a, currentDistance);

                vec3 accumulatedColor   = vec3(0.0);

                float accumulationWeight = (offscreen) ? 1.0 : 1.0 - exp(-max(distanceDelta - 0.1, 0.0) * 2.0);

                if (offscreen) accumulatedColor = currentColor.rgb;
                else accumulatedColor =  mix(historyData.rgb, currentColor.rgb, accumulationWeight * 0.96 + 0.04);

                temporal_data   = clamp_drawbuffer(vec4(accumulatedColor, currentDistance));
                indirectLight_data.rgb  = accumulatedColor;
            #endif
            
            indirectLight_data.a    = currentColor.a;
        }
    #endif

    indirectLight_data  = clamp_drawbuffer(indirectLight_data);
}