#version 450 compatibility

/*
====================================================================================================
    __  __    __  ___    ____    ____ 
    \ \/ /   /  |/  /   /  _/   / __ \
     \  /   / /|_/ /    / /    / /_/ /
     / /   / /  / /   _/ /    / _, _/ 
    /_/   /_/  /_/   /___/   /_/ |_|  

    YMIR SHADERS

    Copyright (C) 2020 Jessie, RRe36, Niemand

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://github.com/ymir-graphics/ymir-releases/blob/master/LICENSE

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/



#include "/lib/universal.glsl"


/* ------ framebuffer attachment outputs ------ */
/*DRAWBUFFERS:034*/
layout(location = 0) out vec4 sceneColor;
layout(location = 1) out vec4 fogScattering;
layout(location = 2) out vec4 fogTransmittance;


/* ------ inputs from vertex stage ------ */
in vec2 coord;

in vec3 lightVector;


/* ------ uniforms ------ */
uniform int frameCounter;
uniform int isEyeInWater;

uniform float eyeAltitude;
uniform float frameTimeCounter;

uniform ivec2 eyeBrightness;

uniform vec2 pixelSize;
uniform vec2 viewSize;
uniform vec2 taaOffset;

uniform vec3 cameraPosition;
uniform vec3 fogData;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 shadowModelView, shadowModelViewInverse;
uniform mat4 shadowProjection, shadowProjectionInverse;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex4;
uniform sampler2D colortex6;

uniform sampler2D depthtex0;
uniform sampler2D depthtex1;

uniform sampler2D shadowtex0;
uniform sampler2D shadowtex1;
uniform sampler2D shadowcolor0;

uniform sampler2D noisetex;
uniform sampler3D depthtex2;


/* ------ includes ------*/
#include "/lib/util/srgb.glsl"
#include "/lib/util/encoders.glsl"
#include "/lib/util/transforms.glsl"
#include "/lib/fragment/dither.glsl"
#include "/lib/lighting/warp.glsl"
#define PD_3DNOISE
#include "/lib/fragment/noise.glsl"
#include "/lib/atmosphere/constants.glsl"



/* ------ functions ------ */
vec3 decode3x8(float a){
    int bf = int(a*65535.);
    return vec3(bf%32, (bf>>5)%64, bf>>11) / vec3(31,63,31);
}


/* ------ shadow position transforms ------ */
vec3 sceneToShadowSpace(vec3 scenePos, const float bias, out float cd) {
    float a     = length(scenePos);
    vec3 pos    = scenePos + bias * lightVector;
        pos     = transMAD(shadowModelView, pos);
        pos     = projMAD3(shadowProjection, pos);
        pos.z  *= 0.2;
        pos.z  -= 0.0012*(saturate(a/256.0));
        float warp = 1.0;
        pos.xy  = warp_shadowmap(pos.xy, warp);

    vec3 temp   = pos;
        temp.xy *= warp;
        temp    = projMAD3(shadowProjectionInverse, temp);
        cd      = -temp.z;

    return pos * 0.5 + 0.5;
}
vec2 cloudShadowSpace(vec3 scenePos) {
    vec3 pos    = mat3(shadowModelView) * scenePos;
        pos    *= rcp(vcloud_shadow_clip);
        pos.xy  = pos.xy * 0.5 + 0.5;

    vec2 mcoeff = viewSize * rcp(1024.0);

    return pos.xy * rcp(mcoeff);
}


/* ------ shadow depth comparision ------*/
float get_shadowtex(sampler2D tex, vec2 coord) {
    return texture(tex, coord).x;
}
float get_shadow(float depth, float cd) {
    return 1.0-saturate((cd-depth*256.0-0.0085)*25.0);
}
float get_shadow(sampler2D tex, vec2 coord, float cd) {
    return get_shadow(get_shadowtex(tex, coord).x, cd);
}
vec3 get_shadowcol(sampler2D tex, vec2 coord) {
    vec4 x  = texture(tex, coord);
    return mix(vec3(1.0), x.rgb, x.a);
}

/* ------ water volume ------ */
vec3 compute_waterVolume(vec3 scenePos0, vec3 scenePos1, float dither) {
    bool isUnderwater   = isEyeInWater == 1;

    vec3 epos   = isUnderwater ? scenePos0 : scenePos1;
    vec3 spos   = isUnderwater ? gbufferModelViewInverse[3].xyz : scenePos0;

    float bstep      = length((epos-spos));
    float stepcoeff  = saturate(bstep * rcp(64.0));
    uint steps       = 4 + uint(stepcoeff * 6);

    vec3 rstep      = (epos-spos)/steps;
    vec3 rpos       = rstep*dither+spos;
    float rlength   = length(rstep);

    vec3 transmittance = vec3(1.0);

    const vec3 absorbCoeff = vec3(2e-1, 4e-2, 1e-2);
    const vec3 scatterCoeff = vec3(1.5e-2);
    const vec3 attenCoeff = (absorbCoeff + scatterCoeff) * 4.0;

    for (uint i = 0; i < steps; i++, rpos += rstep) {
        if (max(transmittance.x, max(transmittance.y, transmittance.z)) < 0.01) break;

        float od    = rlength;
        transmittance  *= expf(-attenCoeff * od);
    }

    return transmittance;
}

vec3 blendTranslucencies(vec3 sceneCol, vec4 translucents, vec3 albedo) {
    vec3 color  = sceneCol;
        color  *= mix(vec3(1.0), albedo, translucents.a);
        color   = mix(color, translucents.rgb, translucents.a);

    return color;
}

void main() {
    vec3 lightVectorView = mat3(gbufferModelView) * lightVector;

    vec4 sceneCol      = stex(colortex0);
    float sceneDepth0  = stex(depthtex0).r;
    float sceneDepth1  = stex(depthtex1).r;

    vec4 tex2          = stex(colortex2);
    int matID          = decodeMatID16(tex2.x);

    float bluenoise    = dither_bluenoise();
    

    vec3 viewPos0      = screenSpaceToViewSpace(vec3(coord, sceneDepth0), gbufferProjectionInverse, true);
    vec3 scenePos0     = viewSpaceToSceneSpace(viewPos0, gbufferModelViewInverse);

    #ifdef vfog_enabled
        vec3 viewPos1  = screenSpaceToViewSpace(vec3(coord, sceneDepth1), gbufferProjectionInverse, true);
        vec3 scenePos1 = viewSpaceToSceneSpace(viewPos1, gbufferModelViewInverse);
        float vDotL    = dot(normalize(viewPos1), lightVectorView);

        vec3 waterVolumeFogTransmittance;

        if (sceneDepth0 < sceneDepth1) {
            vec4 tex1                   = stex(colortex1);
            vec4 translucentColor       = stex(colortex3);
            vec3 translucentAlbedo      = sqr(decode3x8(tex1.b));

            if (matID == 101 && isEyeInWater == 0) {
                waterVolumeFogTransmittance = compute_waterVolume(scenePos0, scenePos1, bluenoise);
                sceneCol.rgb            = sceneCol.rgb * waterVolumeFogTransmittance;
            }

            sceneCol.rgb                = blendTranslucencies(sceneCol.rgb, translucentColor, translucentAlbedo);
        } 
        
        if (isEyeInWater == 1) {
            waterVolumeFogTransmittance = compute_waterVolume(scenePos0, scenePos1, bluenoise);

            sceneCol.rgb                = sceneCol.rgb * waterVolumeFogTransmittance;
        }
    #else
        if (sceneDepth0 < sceneDepth1) {
            vec4 tex1                   = stex(colortex1);
            vec4 translucentColor       = stex(colortex3);
            vec3 translucentAlbedo      = sqr(decode3x8(tex1.b));
            sceneCol.rgb                = blendTranslucencies(sceneCol.rgb, translucentColor, translucentAlbedo);
        }
    #endif

    sceneColor         = make_drawbuffer(sceneCol);
}
